#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2021 CORIA
"""Vector operators (mainly divergence)."""

from typing import Optional

import pandas as pd
import xarray as xr
from fipy import CellVariable, DiffusionTerm
from fipy.meshes.uniformGrid2D import UniformGrid2D
from pint import UnitRegistry


def constant(
    w: float,
    d: float,
    n0: float,
    coord_unit: str,
    input_unit: Optional[str] = None,
    w_unit: Optional[str] = None,
    d_unit: Optional[str] = None,
    verbose: bool = False,
) -> float:
    """BOS Vinnishenko et al."""
    if input_unit is not None:
        if verbose and (w_unit, d_unit) != (None, None):
            print(f"ignoring w_unit and d_unit (using input_unit={input_unit})")
        w_unit = d_unit = input_unit
    else:
        if w_unit is None:
            if verbose:
                print(f"assuming w is in {coord_unit} (the same unit as the coords)")
            w_unit = coord_unit
        if d_unit is None:
            if verbose:
                print(f"assuming d is in {coord_unit} (the same unit as the coords)")
            d_unit = coord_unit

    ureg = UnitRegistry()
    d = d * ureg(d_unit)
    w = w * ureg(w_unit)
    output_unit = 1 / ureg(coord_unit)

    cte = 2 * n0 / (2 * d + w)
    cte = cte.to(output_unit)  # type: ignore

    if verbose:
        print(f"The constant is {cte}")

    return cte.magnitude  # type: ignore


def solve_BOS(
    divergence: xr.DataArray,
    w: float,
    d: float,
    n0: float,
    idz_top: Optional[float] = None,
    idz_bottom: Optional[float] = None,
    idz_left: Optional[float] = None,
    idz_right: Optional[float] = None,
    input_unit: Optional[str] = None,
    w_unit: Optional[str] = None,
    d_unit: Optional[str] = None,
    verbose: bool = False,
) -> xr.DataArray:
    """Solve the diffusion equation lap(n)=div."""
    # pylint: disable=too-many-branches
    dx = float(divergence.x[1] - divergence.x[0])
    dy = float(divergence.y[1] - divergence.y[0])
    mesh = UniformGrid2D(
        dx=dx,
        dy=dy,
        nx=divergence.x.size,
        ny=divergence.y.size,
        origin=(divergence.x[0:1] - dx / 2, divergence.y[0:1] - dx / 2),
    )

    Cst = constant(
        w,
        d,
        n0,
        coord_unit=divergence.x.units,
        input_unit=input_unit,
        w_unit=w_unit,
        d_unit=d_unit,
        verbose=verbose,
    )

    idz = CellVariable(name="idz", mesh=mesh)
    B = CellVariable(name="B", mesh=mesh, value=Cst * divergence.to_numpy().ravel())

    eq = DiffusionTerm() == B

    if idz_top is None:
        if verbose:
            print("The top boundary condition is Neuman")
        idz.faceGrad.constrain(((0,), (0,)), where=mesh.facesTop)
    else:
        if verbose:
            print("The top boundary condition is Dirichlet")
        idz.constrain(idz_top, where=mesh.facesTop)

    if idz_bottom is None:
        if verbose:
            print("The bottom boundary condition is Neuman")
        idz.faceGrad.constrain(((0,), (0,)), where=mesh.facesBottom)
    else:
        if verbose:
            print("The bottom boundary condition is Dirichlet")
        idz.constrain(idz_bottom, where=mesh.facesBottom)

    if idz_left is None:
        if verbose:
            print("The left boundary condition is Neuman")
        idz.faceGrad.constrain(((0,), (0,)), where=mesh.facesLeft)
    else:
        if verbose:
            print("The left boundary condition is Dirichlet")
        idz.constrain(idz_left, where=mesh.facesLeft)

    if idz_right is None:
        if verbose:
            print("The right boundary condition is Neuman")
        idz.faceGrad.constrain(((0,), (0,)), where=mesh.facesRight)
    else:
        if verbose:
            print("The right boundary condition is Dirichlet")
        idz.constrain(idz_right, where=mesh.facesRight)

    eq.solve(var=idz)

    df = pd.DataFrame({"x": mesh.x, "y": mesh.y, "idz": idz}).set_index(["y", "x"])
    idz = df.to_xarray().idz
    assert isinstance(idz, xr.DataArray)

    idz.assign_coords(x=divergence.x, y=divergence.y)
    idz.x.attrs = divergence.x.attrs
    idz.y.attrs = divergence.y.attrs

    idz.attrs["units"] = divergence.x.units
    idz.attrs["long_name"] = "Integral of the refractive index"

    return idz



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2021 CORIA

"""Physical laws."""

import xarray as xr


def linear(source: str = "water_reglin", dn_over_dT: float = -1.01e-4):
    """Linear laws for n(T) and T(n).

    for `dn_over_dT`
     * water_reglin: linear regression for water [4-45°C]
     * paraffin_reglin: linear regression for paraffin C18H38
     * Lipkin: Lipkin&Kurtz 61
    """
    # dn/dT = drho/dT * (n0-1)/rho0, K-1 par gladstone-Dale

    if source == "water_reglin":
        dn_over_dT = -1.01e-4
    elif source == "paraffin_reglin":
        dn_over_dT = -3.89e-4
    elif source == "Lipkin":
        dn_over_dT = -6.9e-04 * 0.6
    elif source == "custom":
        pass
    else:
        raise ValueError(f"dn_over_dT {dn_over_dT} unknown")

    def n_T(T):
        n = dn_over_dT * T

        if isinstance(n, xr.DataArray):
            n.name = "n"
            n.attrs["units"] = "dimensionless"
            n.attrs["long_name"] = "Refractive index"
        return n

    def T_n(n):
        T = -n / dn_over_dT

        if isinstance(n, xr.DataArray):
            T.name = "T"
            T.attrs["units"] = "°C"
            T.attrs["long_name"] = "Temperature"
        return T

    return n_T, T_n


def polynomial():
    """Polynomial laws for n(T) and T(n) (2nd order)."""
    a_n, b_n, c_n = -1.4924302650e-6, -2.7825685771e-5, 1.3340392319
    a_T, b_T, c_T = -1.3644133349e6, 3.6249437078e6, -2.4076210909e6

    def n_T(T):
        n = a_n * (T ** 2) + b_n * T + c_n

        if isinstance(n, xr.DataArray):
            n.name = "n"
            n.attrs["units"] = "dimensionless"
            n.attrs["long_name"] = "Refractive index"
        return n

    def T_n(n):
        T = a_T * (n ** 2) + b_T * n + c_T

        if isinstance(n, xr.DataArray):
            T.name = "T"
            T.attrs["units"] = "°C"
            T.attrs["long_name"] = "Temperature"
        return T

    return n_T, T_n


def gladstone_dale(n: xr.DataArray, G: float = 0.000229) -> xr.DataArray:
    """Gladston-Dale relation to convert refractive index to density."""
    density = (n - 1) / G

    density.name = "density"
    density.attrs["units"] = "Kg/m³"
    density.attrs["long_name"] = "Density"
    return density


def ideal_gaz(density: xr.DataArray, p: float, M: float) -> xr.DataArray:
    """Ideal gaz law to convert density to temperature."""
    R = 8.3144621  # [J/K/mol] Arrhenius constant
    T0 = 273.15  # Kelvin to °C

    T = p * M / (density * R) - T0  # pv=nRT

    T.name = "T"
    T.attrs["units"] = "°C"
    T.attrs["long_name"] = "Temperature"

    return T



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2021 CORIA
"""Read displacement data from Davis or Dantec."""

from pathlib import Path
from typing import Any, Dict, Optional, Tuple, Union

import pandas as pd
import xarray as xr
from pint import UnitRegistry

DAVIS_CSV = {"sep": ";", "usecols": (0, 1, 4, 5), "names": ("x", "y", "U", "V"), "header": 0}
DANTEC_CSV = {
    "sep": ",",
    "usecols": (2, 3, 6, 7),
    "names": ("x", "y", "U", "V"),
    "header": 0,
    "skiprows": 5,
}
DAVIS_DAT = {"sep": " ", "usecols": (0, 1, 4, 5), "names": ("x", "y", "U", "V"), "skiprows": 3}
DANTEC_DAT = {
    "sep": None,
    "usecols": (4, 5, 8, 9),
    "names": ("x", "y", "U", "V"),
    "skiprows": 3,
    "skipfooter": 3,
    "engine": "python",
}


def validate_template(
    template: Dict[str, Any],
    x_col: Optional[int] = None,
    y_col: Optional[int] = None,
    u_col: Optional[int] = None,
    v_col: Optional[int] = None,
    verbose: bool = False,
) -> Dict[str, Any]:
    """Make sure that we have enough information in the template to read the data."""
    template = template.copy()
    default_names = "x", "y", "U", "V"
    default_cols = 0, 1, 2, 3
    assert len(default_names) == len(default_cols)
    warning = False

    if "names" not in template:
        print('You should provide the "names" argument providing the order of columns')
        template["names"] = default_names
        warning = True

    if sorted(template["names"]) != sorted(default_names):
        raise ValueError(f"The colomn names must be exactly {default_names} in any order")

    if "usecols" not in template and None in (x_col, y_col, u_col, v_col):
        print("You should specify which columns to use")
        template["usecols"] = 0, 1, 2, 3
        warning = True

    if len(template["usecols"]) != len(default_cols):
        raise ValueError(
            f"The usecols argument must have {len(default_cols)} values"
            f' corresponding to {template["names"]}'
        )

    if (x_col, y_col, u_col, v_col) != (None, None, None, None):
        cols = list(template["usecols"])
        if x_col is not None:
            cols[template["names"].index("x")] = x_col
        if y_col is not None:
            cols[template["names"].index("y")] = y_col
        if u_col is not None:
            cols[template["names"].index("U")] = u_col
        if v_col is not None:
            cols[template["names"].index("V")] = v_col
        template["usecols"] = tuple(cols)

    if warning:
        print(f'Assuming names={template["names"]} and usecols={template["usecols"]}, i.e. ')

    if warning or verbose:
        for name, col in zip(template["names"], template["usecols"]):
            print(f"The data for {name} is expected to be in columns {col}")

    return template


def validate_scale(
    scale: float = 1.0,
    coord_scale: float = 1.0,
    displacement_scale: float = 1.0,
    unit: Optional[str] = None,
    coord_unit: Optional[str] = None,
    displacement_unit: Optional[str] = None,
    output_unit: Optional[str] = None,
    verbose: bool = False,
) -> Tuple[float, float, str]:
    """Make sure that we have enough information in the template to read the data."""
    if unit is not None:
        if verbose and (coord_unit, displacement_unit) != (None, None):
            print(f"ignoring coord_unit and displacement_unit (using unit={unit})")
        coord_unit = displacement_unit = unit
    elif None in (coord_unit, displacement_unit):
        raise ValueError("You must provide the unit of the input data (after the rescaling)")

    if output_unit is None:
        if verbose:
            print(f"The output unit is chosen to be the same as the coord unit ({coord_unit})")
        output_unit = coord_unit
    assert isinstance(output_unit, str)

    if scale is not None:
        if verbose and (coord_scale, displacement_scale) != (None, None):
            print(f"ignoring coord_scale and displacement_scale (using scale={scale})")
        coord_scale = displacement_scale = scale

    ureg = UnitRegistry()
    output_unit_ = ureg(output_unit)
    coord_unit_ = ureg(coord_unit)
    displacement_unit_ = ureg(displacement_unit)

    coord_scale *= coord_unit_.to(output_unit_).magnitude
    displacement_scale *= displacement_unit_.to(output_unit_).magnitude

    if verbose:
        print(
            f"The input data will be scaled to {output_unit_.units}"
            f"with the following scaling factor: "
        )
        print(f"  coord_scale={coord_scale}")
        print(f"  displacement_scale={displacement_scale}")

    return coord_scale, displacement_scale, output_unit


def reader(
    filename: Union[str, Path],
    template: Dict[str, Any],
    scale: float = 1.0,
    coord_scale: float = 1.0,
    displacement_scale: float = 1.0,
    x_col: Optional[int] = None,
    y_col: Optional[int] = None,
    u_col: Optional[int] = None,
    v_col: Optional[int] = None,
    output_unit: Optional[str] = None,
    unit: Optional[str] = None,
    coord_unit: Optional[str] = None,
    displacement_unit: Optional[str] = None,
    verbose: bool = False,
    **kwargs: Any,
) -> xr.Dataset:
    """Read a csv file from Dantec or Davis."""
    filename = Path(filename)
    template = validate_template({**template, **kwargs}, x_col, y_col, u_col, v_col, verbose)
    coord_scale, displacement_scale, unit = validate_scale(
        scale,
        coord_scale,
        displacement_scale,
        unit,
        coord_unit,
        displacement_unit,
        output_unit,
        verbose,
    )

    dataframe = pd.read_csv(filename, **template)

    dataframe.x *= coord_scale
    dataframe.y *= coord_scale

    dataframe.U *= displacement_scale
    dataframe.V *= displacement_scale

    dataframe = dataframe.set_index(["y", "x"])
    dataset = dataframe.to_xarray()
    assert isinstance(dataset, xr.Dataset)

    dataset.x.attrs["units"] = unit
    dataset.y.attrs["units"] = unit
    dataset.U.attrs["units"] = unit
    dataset.V.attrs["units"] = unit
    dataset.U.attrs["long_name"] = "Displacement along x"
    dataset.V.attrs["long_name"] = "Displacement along y"

    return dataset


if __name__ == "__main__":
    data = reader(
        "BOS_tanguy_15092021/test1_Iminsta0_cube_eau_2607_PIV_Imref0_9Im_.6o5xcjgo.000000.dat",
        template=DANTEC_DAT,
        scale=2,
        unit="mm",
        output_unit="m",
        verbose=True,
    )
    print(data)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2021 CORIA
"""Demonstration of a BOS script."""

from pathlib import Path

import matplotlib.pyplot as plt
import pandas as pd
import xarray as xr
from scipy.signal import spline_filter

from pybos.extraction import idz2n
from pybos.laplacian import solve_BOS
from pybos.physical_laws import gladstone_dale, ideal_gaz
from pybos.reader import DANTEC_DAT, reader
from pybos.vector_calculus import divergence

pybos_path = test_path = Path(__file__).parent
while pybos_path.name != "tests":
    pybos_path = pybos_path.parent
pybos_path = pybos_path.parent


def filter_(x: float) -> float:
    """Apply a gaussian filter."""
    return spline_filter(x, lmbda=2)


def read_data() -> xr.Dataset:
    """Read example data."""
    im_datafile = pybos_path / "data/example_data_axi.vect"
    cols = 0, 1, 2, 3  # pix for now
    scale = 1 / 77.0  # pix to mm

    data = reader(
        im_datafile,
        DANTEC_DAT,
        usecols=cols,
        scale=scale,
        unit="mm",
        output_unit="mm",
        skipfooter=0,
    )
    return data


def crop(data: xr.Dataset) -> xr.Dataset:
    """Crop data."""
    roi = {"x": slice(None, None), "y": slice(20, -20)}

    cropped_data = data.isel(roi)  # type: ignore
    return cropped_data


def plot(T: xr.DataArray) -> None:
    """Plot the resulting temperature."""
    fig, ax = plt.subplots()
    T.plot.pcolormesh(ax=ax, cmap=plt.cm.jet, vmin=20, vmax=90)
    ax.axis("equal")
    fig.tight_layout()
    fig.savefig(pybos_path / "artifacts/Temperature_axi.png", dpi=200)


def check(**res: float) -> None:
    """Check some reference values."""
    ref = pd.Series(
        {
            "divmin": -0.014772756879968412,
            "divmean": 0.001903474952733359,
            "divmax": 0.02153539914658238,
            "Tmin": 26.74070754311657,
            "Tmean": 52.11628994045219,
            "Tmax": 89.0648083252903,
        }
    )
    pd.testing.assert_series_equal(pd.Series(res).astype(float), ref)


def test_axi() -> None:
    """Run an axi case."""
    d = 55  # [mm] object-pattern distance
    n0 = 1.00027  # [] refractive index of the air
    G = 0.000226  # [m³/Kg] Gladstone constant of air at 20°C and 1 bar
    p = 101_325  # [Pa] = [J/m³] pressure
    M = 28.976 * 1e-3  # [kg/mol] molar mass

    data = read_data()
    cropped_data = crop(data)

    filtered_data = xr.apply_ufunc(filter_, cropped_data, keep_attrs=True)
    filtered_data["divergence"] = divergence(filtered_data)

    w = float(filtered_data.x.max() - filtered_data.x.min())  # [mm] object size

    idz_top = idz_bottom = None
    idz_left = n0 * w
    idz_right = n0 * w

    idz = solve_BOS(filtered_data.divergence, w, d, n0, idz_top, idz_bottom, idz_left, idz_right)

    # using an axisymmetric hypothesis
    n = idz2n(idz, w, n0, method="abel")

    density = gladstone_dale(n, G)
    T = ideal_gaz(density, p, M)

    plot(T)

    check(
        divmin=filtered_data.divergence.min(),
        divmean=filtered_data.divergence.mean(),
        divmax=filtered_data.divergence.max(),
        Tmin=T.min(),
        Tmean=T.mean(),
        Tmax=T.max(),
    )


if __name__ == "__main__":
    test_axi()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2021 CORIA
"""Demonstration of a BOS script."""

from pathlib import Path

import matplotlib.pyplot as plt
import pandas as pd
import xarray as xr
from scipy.ndimage import gaussian_filter

from pybos.extraction import idz2n
from pybos.laplacian import solve_BOS
from pybos.physical_laws import polynomial
from pybos.reader import DANTEC_DAT, reader
from pybos.vector_calculus import divergence

pybos_path = test_path = Path(__file__).parent
while pybos_path.name != "tests":
    pybos_path = pybos_path.parent
pybos_path = pybos_path.parent


def filter_(x: float) -> float:
    """Apply a gaussian filter."""
    return gaussian_filter(x, sigma=(2, 2), order=0, mode="nearest")


def read_data() -> xr.Dataset:
    """Read example data."""
    im_datafile = pybos_path / "data/example_data_2d.dat"
    cols = 2, 3, 6, 7  # pix for now
    scale = 1 / (2560 * ((283 - 26) / 319) / 384)  # pix to mm

    data = reader(im_datafile, DANTEC_DAT, usecols=cols, scale=scale, unit="mm")
    data["U"] *= -1
    data["V"] *= -1
    return data


def crop(data: xr.Dataset) -> xr.Dataset:
    """Crop data."""
    pix1, piy1 = 244, 117
    pix2, piy2 = 2321, 2012

    Pix_size = 8

    pix1 = int((pix1 - Pix_size + 0.5) / Pix_size)
    pix2 = int((pix2 - Pix_size + 0.5) / Pix_size)
    piy1 = int((piy1 - Pix_size + 0.5) / Pix_size)
    piy2 = int((piy2 - Pix_size + 0.5) / Pix_size)

    piy1, piy2 = data.dims["y"] - piy2, data.dims["y"] - piy1

    roi = {"x": slice(pix1, pix2), "y": slice(piy1, piy2)}
    cropped_data = data.isel(roi)  # type: ignore
    return cropped_data


def plot(T: xr.DataArray) -> None:
    """Plot the resulting temperature."""
    fig, ax = plt.subplots()
    T.plot.pcolormesh(ax=ax, cmap=plt.cm.jet, vmin=0, vmax=50)
    ax.axis("equal")
    fig.tight_layout()
    fig.savefig(pybos_path / "artifacts/Temperature_2d.png", dpi=200)


def check(**res: float) -> None:
    """Check some reference values."""
    ref = pd.Series(
        {
            "divmin": -0.11082972510245184,
            "divmean": 0.0009025630571096038,
            "divmax": 0.1762545837433431,
            "Tmin": 7.173597980756313,
            "Tmean": 26.68481712640316,
            "Tmax": 41.65479086851701,
        }
    )
    pd.testing.assert_series_equal(pd.Series(res).astype(float), ref)


def test_2d() -> None:
    """Run a 2d case."""
    d = 190  # for cube # 0.003 for aquarium # [mm] pattern-tank distance
    w = 400  # [mm] tank width "h" in [Plakina et al 2012]
    n0 = 1.00027  # [] refractive index of the air

    data = read_data()
    cropped_data = crop(data)

    filtered_data = xr.apply_ufunc(filter_, cropped_data, keep_attrs=True)
    filtered_data["divergence"] = divergence(filtered_data)

    n_T, T_n = polynomial()

    T_left = 7.5
    T_right = 39.5

    idz_top = idz_bottom = None
    idz_left = n_T(T_left) * w
    idz_right = n_T(T_right) * w

    idz = solve_BOS(filtered_data.divergence, w, d, n0, idz_top, idz_bottom, idz_left, idz_right)

    # using a 2d hypothesis
    n = idz2n(idz, w, method="2d")

    T = T_n(n)
    plot(T)

    check(
        divmin=filtered_data.divergence.min(),
        divmean=filtered_data.divergence.mean(),
        divmax=filtered_data.divergence.max(),
        Tmin=T.min(),
        Tmean=T.mean(),
        Tmax=T.max(),
    )


if __name__ == "__main__":
    test_2d()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2021 CORIA
"""pyBOS demonstration script.

This script is not intented to only be useful for pedagogical purposes.
It is not either a lecture about BOS but only a demonstration
of what can be done with the library pyBOS.
"""


# First we import packages that we'll use later.
from pathlib import Path

import matplotlib.pyplot as plt
import xarray as xr

# Then we import stuff fron the pyBOS library
from pybos.extraction import idz2n
from pybos.laplacian import solve_BOS
from pybos.physical_laws import polynomial
from pybos.reader import DANTEC_DAT, reader
from pybos.vector_calculus import amplitude, divergence

# We define some of the parameter that we'll need later

fig_show = 1  # plot only the temperature
# fig_show = 2 # plot the temperature and U, V, norm and the divergence
# fig_show = 3 # plot all

reverse = True  # if reference image and final image are reversed (should not happen !)
crop = True
filt_method = "gaussian"

"################################### READING THE DATA #############################################"
# We use the `reader` function to read the data.
# It must be a `.csv` type file (including dome of the tecplot's `.plt.` files)
#
# If possible prefer using the calibration from Dantec/Davis and avoid importing data in pixel.
im_datafile = Path("data/example_data_2d.dat")
cols = 2, 3, 6, 7  # pix for now
scale = 1 / (2560 * ((283 - 26) / 319) / 384)  # pix to mm

data = reader(im_datafile, DANTEC_DAT, usecols=cols, scale=scale, unit="mm")
if reverse:
    data["U"] *= -1
    data["V"] *= -1

if fig_show >= 3:
    fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(9, 3))
    data.U.plot.pcolormesh(ax=ax1, cmap=plt.cm.jet)
    data.V.plot.pcolormesh(ax=ax2, cmap=plt.cm.jet)
    ax1.axis("equal")
    ax2.axis("equal")
    fig.tight_layout()
    plt.show()

"################################## PREPARING THE DATA ############################################"
# You may want to crop the data to remove reflection and other defects.
# Here we are only using the `isel` method of `Xarray`'s `Dataset`.
roi = {"x": slice(None, None), "y": slice(None, None)}
if crop:
    pix1, piy1 = 244, 117
    pix2, piy2 = 2321, 2012
    Pix_size = 8

    pix1 = int((pix1 - Pix_size + 0.5) / Pix_size)
    pix2 = int((pix2 - Pix_size + 0.5) / Pix_size)
    piy1 = int((piy1 - Pix_size + 0.5) / Pix_size)
    piy2 = int((piy2 - Pix_size + 0.5) / Pix_size)

    piy1, piy2 = data.dims["y"] - piy2, data.dims["y"] - piy1

    roi = {"x": slice(pix1, pix2), "y": slice(piy1, piy2)}
cropped_data = data.isel(roi)  # type: ignore

if fig_show >= 3:
    fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(9, 3))
    data.U.plot.pcolormesh(ax=ax1, cmap=plt.cm.jet)
    (cropped_data.U * 0.0).plot.pcolormesh(ax=ax1, alpha=0.5, add_colorbar=False)
    data.V.plot.pcolormesh(ax=ax2, cmap=plt.cm.jet)
    (cropped_data.V * 0.0).plot.pcolormesh(ax=ax2, alpha=0.5, add_colorbar=False)
    ax1.axis("equal")
    ax2.axis("equal")
    fig.tight_layout()
    plt.show()

# You may also want to filter the data. Here we are only using standard filter from `scipy`.
# In order to keep the `Xarray`'s `Dataset` data format, we use the `apply_ufunc` wrapper.
if filt_method == "gaussian":
    from scipy.ndimage import gaussian_filter

    sigma = 2, 2

    def filter_(x: float) -> float:
        """Apply a gaussian filter."""
        return gaussian_filter(x, sigma=sigma, order=0, mode="nearest")


elif filt_method == "spline":
    from scipy.signal import spline_filter

    lmbda = 2

    def filter_(x: float) -> float:
        """Apply a spline filter."""
        return spline_filter(x, lmbda=lmbda)


filtered_data = xr.apply_ufunc(filter_, cropped_data, keep_attrs=True)

# There is a function for computing the displacement's amplitude and divergence.
filtered_data["divergence"] = divergence(filtered_data)
print(f"mean(divergence)={float(filtered_data.divergence.mean())}")

if fig_show >= 2:
    filtered_data["amplitude"] = amplitude(filtered_data)

    fig, axs = plt.subplots(2, 2, figsize=(9, 9))
    filtered_data.U.plot.pcolormesh(ax=axs[0][0], cmap=plt.cm.jet)
    filtered_data.V.plot.pcolormesh(ax=axs[0][1], cmap=plt.cm.jet)
    filtered_data.amplitude.plot.pcolormesh(ax=axs[1][0], cmap=plt.cm.jet)
    filtered_data.divergence.plot.pcolormesh(ax=axs[1][1], cmap=plt.cm.jet)

    for axx in axs:
        for axy in axx:
            axy.axis("equal")
    fig.tight_layout()
    plt.show()

"######################################## Resolution ##############################################"
# Now we define the properties of the experiment (size of the tank, boundary conditions)
# and the physical law for converting refracting index to to temperature and conversely.

d = 190  # for cube # 0.003 for aquarium # [mm] pattern-tank distance
w = 400  # [mm] tank width "h" in [Plakina et al 2012]
n0 = 1.00027  # [] refractive index of the air

n_T, T_n = polynomial()

T_left = 7.5
T_right = 39.5

idz_top = idz_bottom = None
idz_left = n_T(T_left) * w
idz_right = n_T(T_right) * w

# And compute the refractive index

idz = solve_BOS(filtered_data.divergence, w, d, n0, idz_top, idz_bottom, idz_left, idz_right)
res = idz.to_dataset()

# using a 2d hypothesis
res["n"] = idz2n(res.idz, w, method="2d")

if fig_show >= 3:
    fig, ax = plt.subplots()
    res.n.plot.pcolormesh(ax=ax, cmap=plt.cm.jet)
    ax.axis("equal")
    fig.tight_layout()
    plt.show()

# That we can convert to temperature using the aformentionned law
res["T"] = T_n(res.n)

fig, ax = plt.subplots()
res.T.plot.pcolormesh(ax=ax, cmap=plt.cm.jet, vmin=0, vmax=50)
ax.axis("equal")
fig.tight_layout()
fig.savefig("Temperature_2d.png", dpi=200)

if fig_show >= 1:
    plt.show()

# Now it is your turn to use the library ;) !



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2021 CORIA
"""Radon and Abel transformation."""
from typing import Any, Optional

import numpy as np
import xarray as xr
from pint import UnitRegistry

try:
    from skimage.transform import iradon
except ImportError:
    with_iradon = False
else:
    with_iradon = True

try:
    from abel import Transform
except ImportError:
    with_abel = False
else:
    with_abel = True


def _radon(
    idz: xr.DataArray, angular_resolution: int = 180, filter_name: str = "hamming"
) -> xr.DataArray:
    """Compute the refractive index using a Radon transform."""
    if not with_iradon:
        raise ImportError("Please install scikit-image to access the radon transform.")

    def transform(x: np.ndarray) -> np.ndarray:
        """Compute a radon transform assuming that the signal is constant wrt the angle."""
        II = np.tile(x, (angular_resolution, 1)).T
        res = iradon(II, output_size=x.size, filter_name=filter_name)
        return res

    n = xr.apply_ufunc(
        transform, idz, input_core_dims=[["x"]], output_core_dims=[["z", "x"]], vectorize=True
    )
    n = n.isel({"z": n.z.size // 2})

    return n


def _abel(idz: xr.DataArray, **kwargs: Any) -> xr.DataArray:
    """Compute the refractive index using a Abel transform."""
    if not with_abel:
        raise ImportError("Please install PyAbel to access the abel transform.")

    # defaults values
    kwargs["origin"] = kwargs.get("origin", "image_center")
    kwargs["symmetry_axis"] = kwargs.get("symmetry_axis", 0)
    kwargs["center_options"] = kwargs.get("center_options", {"crop": "maintain_size"})

    def transform(x: np.ndarray) -> np.ndarray:
        """Compute an abel transform."""
        return Transform(x, direction="backward", method="hansenlaw", **kwargs).transform

    n = xr.apply_ufunc(transform, idz)

    return n


def idz2n(
    idz: xr.DataArray,
    w: float,
    n0: float = 1.0,
    method: str = "2d",
    w_unit: Optional[str] = None,
    verbose: bool = False,
    **kwargs: Any,
) -> xr.DataArray:
    """Extract the refractive index from its integral using a 2d or axysimmetric hypothesis."""
    if w_unit is not None:
        if verbose:
            print(f"Scaling w from {w_unit} to {idz.x.units}")
        ureg = UnitRegistry()
        w = (w * ureg(w_unit)).to(ureg(idz.x.units)).magnitude
    elif verbose:
        print("Assuming that w and coords are in the same unit.")

    if method == "2d":
        n = idz / w
    elif method == "radon":
        n = _radon(idz - n0 * w, **kwargs) + n0
    elif method == "abel":
        n = _abel(idz - n0 * w, **kwargs) + n0
    else:
        raise ValueError(f"{method} method unknown")

    n.name = "n"
    n.attrs["units"] = "dimensionless"
    n.attrs["long_name"] = "Refractive index"

    return n



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2021 CORIA
"""Vector operators (mainly divergence)."""

import xarray as xr


def divergence(
    data: xr.Dataset,
    edge_order: int = 0,
    mask: bool = True,
    U: str = "U",
    V: str = "V",
    verbose: bool = False,
) -> xr.DataArray:
    """Compute the divergence of a 2D vector field."""
    edge_order_ = max(edge_order, 1)
    grad_x = data[U].differentiate("x", edge_order=edge_order_)
    grad_y = data[V].differentiate("y", edge_order=edge_order_)

    if not edge_order:
        grad_x[{"x": [0, -1]}] = 0
        grad_y[{"y": [0, -1]}] = 0

    div = grad_x + grad_y

    if mask:
        div = div.where(abs(data[U]) > 1e-30, 0.0)

    div.name = "div"
    div.attrs["units"] = "dimensionless"
    div.attrs["long_name"] = "Displacement divergence"

    if verbose:
        print(f"The divergence is computed as : div = d({U})/dx + d({V})/dx")
        print(f"Using a {edge_order}-th order accurate differences at the boundaries.")
        if not edge_order:
            print(" (the normal derivative is taken to be 0 at the boundaries)")

        if mask:
            print(f"A mask is applied to remove the values where |{U}|<=1e-30")

        print(f"mean(divergence)={float(div.mean())}")

    return div


def amplitude(data: xr.Dataset, U: str = "U", V: str = "V", verbose: bool = False) -> xr.DataArray:
    """Compute the amplitude of a 2D vector field."""
    norm = xr.ufuncs.hypot(data[U], data[V])  # type: ignore # pylint: disable=no-member
    assert isinstance(norm, xr.DataArray)

    norm.name = "norm"
    norm.attrs["units"] = data[U].units
    norm.attrs["long_name"] = "Displacement amplitude"

    if verbose:
        print(f"mean(amplitude)={float(norm.mean())}")

    return norm



#!/usr/bin/env python3
# coding=utf-8
# Copyright (C) 2021 CORIA
"""Compile and install pyBOS."""
from subprocess import CalledProcessError, run

from setuptools import setup

try:
    run(["git", "branch"], check=True)
except (CalledProcessError, FileNotFoundError):
    setup()
else:
    # Add the version from git
    # --------------------------------
    setup(use_scm_version={"write_to": "pybos/version.py", "fallback_version": "0.0.0"})
